@@ -8,7 +8,8 @@ use netidx::{
     path::Path,
     subscriber::{self, Dval, Typ, UpdatesFlags, Value},
 };
-use std::{collections::HashSet, marker::PhantomData, sync::Arc};
+use netidx_core::utils::Either;
+use std::{collections::HashSet, iter, marker::PhantomData, sync::Arc};
 
 pub struct CachedVals(pub Vec<Option<Value>>);
 
@@ -33,13 +34,20 @@ impl CachedVals {
             }
         })
     }
+
+    pub fn flat_iter<'a>(&'a self) -> impl Iterator<Item = Option<Value>> + 'a {
+        self.0.iter().flat_map(|v| match v {
+            None => Either::Left(iter::once(None)),
+            Some(v) => Either::Right(v.clone().flatten().map(Some)),
+        })
+    }
 }
 
 pub struct Any(Option<Value>);
 
 impl<C: Ctx, E> Register<C, E> for Any {
     fn register(ctx: &mut ExecCtx<C, E>) {
-        let f: InitFn<C, E> = Arc::new(|_ctx, from, _| {
+        let f: InitFn<C, E> = Arc::new(|_ctx, from, _, _| {
             Box::new(Any(from.iter().find_map(|s| s.current())))
         });
         ctx.functions.insert("any".into(), f);
@@ -73,7 +81,7 @@ pub struct Do(Option<Value>);
 
 impl<C: Ctx, E> Register<C, E> for Do {
     fn register(ctx: &mut ExecCtx<C, E>) {
-        let f: InitFn<C, E> = Arc::new(|_ctx, from, _| {
+        let f: InitFn<C, E> = Arc::new(|_ctx, from, _, _| {
             Box::new(Do(from.iter().fold(None, |_, s| s.current())))
         });
         ctx.functions.insert("do".into(), f);
@@ -109,7 +117,7 @@ pub struct CachedCur<T: CachedCurEval> {
 
 impl<C: Ctx, E, T: CachedCurEval + 'static> Register<C, E> for CachedCur<T> {
     fn register(ctx: &mut ExecCtx<C, E>) {
-        let f: InitFn<C, E> = Arc::new(|_ctx, from, _| {
+        let f: InitFn<C, E> = Arc::new(|_ctx, from, _, _| {
             let cached = CachedVals::new(from);
             let current = T::eval(&cached);
             Box::new(CachedCur::<T> { cached, current, t: PhantomData })
@@ -178,7 +186,7 @@ pub struct SumEv;
 
 impl CachedCurEval for SumEv {
     fn eval(from: &CachedVals) -> Option<Value> {
-        from.0.iter().fold(None, |res, v| match res {
+        from.flat_iter().fold(None, |res, v| match res {
             res @ Some(Value::Error(_)) => res,
             res => add_vals(res, v.clone()),
         })
@@ -204,7 +212,7 @@ fn prod_vals(lhs: Option<Value>, rhs: Option<Value>) -> Option<Value> {
 
 impl CachedCurEval for ProductEv {
     fn eval(from: &CachedVals) -> Option<Value> {
-        from.0.iter().fold(None, |res, v| match res {
+        from.flat_iter().fold(None, |res, v| match res {
             res @ Some(Value::Error(_)) => res,
             res => prod_vals(res, v.clone()),
         })
@@ -230,7 +238,7 @@ fn div_vals(lhs: Option<Value>, rhs: Option<Value>) -> Option<Value> {
 
 impl CachedCurEval for DivideEv {
     fn eval(from: &CachedVals) -> Option<Value> {
-        from.0.iter().fold(None, |res, v| match res {
+        from.flat_iter().fold(None, |res, v| match res {
             res @ Some(Value::Error(_)) => res,
             res => div_vals(res, v.clone()),
         })
@@ -247,7 +255,7 @@ pub struct MinEv;
 
 impl CachedCurEval for MinEv {
     fn eval(from: &CachedVals) -> Option<Value> {
-        from.0.iter().filter_map(|v| v.clone()).fold(None, |res, v| match res {
+        from.flat_iter().filter_map(|v| v).fold(None, |res, v| match res {
             None => Some(v),
             Some(v0) => {
                 if v < v0 {
@@ -270,7 +278,7 @@ pub struct MaxEv;
 
 impl CachedCurEval for MaxEv {
     fn eval(from: &CachedVals) -> Option<Value> {
-        from.0.iter().filter_map(|v| v.clone()).fold(None, |res, v| match res {
+        from.flat_iter().filter_map(|v| v).fold(None, |res, v| match res {
             None => Some(v),
             Some(v0) => {
                 if v > v0 {
@@ -293,7 +301,7 @@ pub struct AndEv;
 
 impl CachedCurEval for AndEv {
     fn eval(from: &CachedVals) -> Option<Value> {
-        let res = from.0.iter().all(|v| match v {
+        let res = from.flat_iter().all(|v| match v {
             Some(Value::True) => true,
             _ => false,
         });
@@ -315,7 +323,7 @@ pub struct OrEv;
 
 impl CachedCurEval for OrEv {
     fn eval(from: &CachedVals) -> Option<Value> {
-        let res = from.0.iter().any(|v| match v {
+        let res = from.flat_iter().any(|v| match v {
             Some(Value::True) => true,
             _ => false,
         });
@@ -350,7 +358,6 @@ impl CachedCurEval for NotEv {
 
 pub type Not = CachedCur<NotEv>;
 
-// CR estokes: document
 pub struct IsErrEv;
 
 impl CachedCurEval for IsErrEv {
@@ -544,7 +551,9 @@ impl CachedCurEval for ReplaceEv {
     fn eval(from: &CachedVals) -> Option<Value> {
         match &*from.0 {
             [Some(Value::String(pat)), Some(Value::String(rep)), Some(Value::String(val))] => {
-                Some(Value::String(Chars::from(String::from(val.replace(&**pat, &**rep)))))
+                Some(Value::String(Chars::from(String::from(
+                    val.replace(&**pat, &**rep),
+                ))))
             }
             _ => Some(Value::Error(Chars::from("replace expected 3 arguments"))),
         }
@@ -559,50 +568,6 @@ pub type Replace = CachedCur<ReplaceEv>;
 
 pub struct CmpEv;
 
-fn eval_op<T: PartialEq + PartialOrd>(op: &str, v0: T, v1: T) -> Value {
-    match op {
-        "eq" => {
-            if v0 == v1 {
-                Value::True
-            } else {
-                Value::False
-            }
-        }
-        "lt" => {
-            if v0 < v1 {
-                Value::True
-            } else {
-                Value::False
-            }
-        }
-        "gt" => {
-            if v0 > v1 {
-                Value::True
-            } else {
-                Value::False
-            }
-        }
-        "lte" => {
-            if v0 <= v1 {
-                Value::True
-            } else {
-                Value::False
-            }
-        }
-        "gte" => {
-            if v0 >= v1 {
-                Value::True
-            } else {
-                Value::False
-            }
-        }
-        op => Value::Error(Chars::from(format!(
-            "invalid op {}, expected eq, lt, gt, lte, or gte",
-            op
-        ))),
-    }
-}
-
 impl CachedCurEval for CmpEv {
     fn name() -> &'static str {
         "cmp"
@@ -616,42 +581,47 @@ impl CachedCurEval for CmpEv {
                     (None, None) => Some(Value::False),
                     (_, None) => Some(Value::False),
                     (None, _) => Some(Value::False),
-                    (Some(v0), Some(v1)) => match (v0, v1) {
-                        (Value::U32(v0), Value::U32(v1)) => Some(eval_op(&*op, v0, v1)),
-                        (Value::U32(v0), Value::V32(v1)) => Some(eval_op(&*op, v0, v1)),
-                        (Value::V32(v0), Value::V32(v1)) => Some(eval_op(&*op, v0, v1)),
-                        (Value::V32(v0), Value::U32(v1)) => Some(eval_op(&*op, v0, v1)),
-                        (Value::I32(v0), Value::I32(v1)) => Some(eval_op(&*op, v0, v1)),
-                        (Value::I32(v0), Value::Z32(v1)) => Some(eval_op(&*op, v0, v1)),
-                        (Value::Z32(v0), Value::Z32(v1)) => Some(eval_op(&*op, v0, v1)),
-                        (Value::Z32(v0), Value::I32(v1)) => Some(eval_op(&*op, v0, v1)),
-                        (Value::U64(v0), Value::U64(v1)) => Some(eval_op(&*op, v0, v1)),
-                        (Value::U64(v0), Value::V64(v1)) => Some(eval_op(&*op, v0, v1)),
-                        (Value::V64(v0), Value::V64(v1)) => Some(eval_op(&*op, v0, v1)),
-                        (Value::V64(v0), Value::U64(v1)) => Some(eval_op(&*op, v0, v1)),
-                        (Value::I64(v0), Value::I64(v1)) => Some(eval_op(&*op, v0, v1)),
-                        (Value::I64(v0), Value::Z64(v1)) => Some(eval_op(&*op, v0, v1)),
-                        (Value::Z64(v0), Value::Z64(v1)) => Some(eval_op(&*op, v0, v1)),
-                        (Value::Z64(v0), Value::I64(v1)) => Some(eval_op(&*op, v0, v1)),
-                        (Value::F32(v0), Value::F32(v1)) => Some(eval_op(&*op, v0, v1)),
-                        (Value::F64(v0), Value::F64(v1)) => Some(eval_op(&*op, v0, v1)),
-                        (Value::String(v0), Value::String(v1)) => {
-                            Some(eval_op(&*op, v0, v1))
+                    (Some(v0), Some(v1)) => Some(match &**op {
+                        "eq" => {
+                            if v0 == v1 {
+                                Value::True
+                            } else {
+                                Value::False
+                            }
                         }
-                        (Value::Bytes(v0), Value::Bytes(v1)) => {
-                            Some(eval_op(&*op, v0, v1))
+                        "lt" => {
+                            if v0 < v1 {
+                                Value::True
+                            } else {
+                                Value::False
+                            }
                         }
-                        (Value::True, Value::True) => Some(eval_op(&*op, true, true)),
-                        (Value::True, Value::False) => Some(eval_op(&*op, true, false)),
-                        (Value::False, Value::True) => Some(eval_op(&*op, false, true)),
-                        (Value::False, Value::False) => Some(eval_op(&*op, false, false)),
-                        (Value::Ok, Value::Ok) => Some(eval_op(&*op, true, true)),
-                        (Value::Error(v0), Value::Error(v1)) => {
-                            Some(eval_op(&*op, v0, v1))
+                        "gt" => {
+                            if v0 > v1 {
+                                Value::True
+                            } else {
+                                Value::False
+                            }
                         }
-                        (Value::Null, Value::Null) => Some(Value::True),
-                        (_, _) => Some(Value::False),
-                    },
+                        "lte" => {
+                            if v0 <= v1 {
+                                Value::True
+                            } else {
+                                Value::False
+                            }
+                        }
+                        "gte" => {
+                            if v0 >= v1 {
+                                Value::True
+                            } else {
+                                Value::False
+                            }
+                        }
+                        op => Value::Error(Chars::from(format!(
+                            "invalid op {}, expected eq, lt, gt, lte, or gte",
+                            op
+                        ))),
+                    }),
                 },
                 Some(_) => Some(Value::Error(Chars::from(
                     "cmp(op, v0, v1): expected op to be a string",
@@ -857,6 +827,7 @@ pub type StringConcat = CachedCur<StringConcatEv>;
 pub struct Eval<C: Ctx, E> {
     cached: CachedVals,
     current: Result<Node<C, E>, Value>,
+    scope: Path,
 }
 
 impl<C: Ctx, E> Eval<C, E> {
@@ -865,7 +836,7 @@ impl<C: Ctx, E> Eval<C, E> {
             [None] => Err(Value::Null),
             [Some(v)] => match v {
                 Value::String(s) => match s.parse::<Expr>() {
-                    Ok(spec) => Ok(Node::compile(ctx, spec)),
+                    Ok(spec) => Ok(Node::compile(ctx, self.scope.clone(), spec)),
                     Err(e) => {
                         let e = format!("eval(src), error parsing formula {}, {}", s, e);
                         Err(Value::Error(Chars::from(e)))
@@ -883,8 +854,9 @@ impl<C: Ctx, E> Eval<C, E> {
 
 impl<C: Ctx, E> Register<C, E> for Eval<C, E> {
     fn register(ctx: &mut ExecCtx<C, E>) {
-        let f: InitFn<C, E> = Arc::new(|ctx, from, _| {
-            let mut t = Eval { cached: CachedVals::new(from), current: Err(Value::Null) };
+        let f: InitFn<C, E> = Arc::new(|ctx, from, scope, _| {
+            let mut t =
+                Eval { cached: CachedVals::new(from), current: Err(Value::Null), scope };
             t.compile(ctx);
             Box::new(t)
         });
@@ -923,7 +895,7 @@ pub struct Count {
 
 impl<C: Ctx, E> Register<C, E> for Count {
     fn register(ctx: &mut ExecCtx<C, E>) {
-        let f: InitFn<C, E> = Arc::new(|_, from, _| {
+        let f: InitFn<C, E> = Arc::new(|_, from, _, _| {
             Box::new(Count { from: CachedVals::new(from), count: 0 })
         });
         ctx.functions.insert("count".into(), f);
@@ -960,7 +932,7 @@ pub struct Sample {
 
 impl<C: Ctx, E> Register<C, E> for Sample {
     fn register(ctx: &mut ExecCtx<C, E>) {
-        let f: InitFn<C, E> = Arc::new(|_, from, _| {
+        let f: InitFn<C, E> = Arc::new(|_, from, _, _| {
             let current = match from {
                 [trigger, source] => match trigger.current() {
                     None => None,
@@ -1017,7 +989,7 @@ pub struct Mean {
 
 impl<C: Ctx, E> Register<C, E> for Mean {
     fn register(ctx: &mut ExecCtx<C, E>) {
-        let f: InitFn<C, E> = Arc::new(|_, from, _| {
+        let f: InitFn<C, E> = Arc::new(|_, from, _, _| {
             Box::new(Mean { from: CachedVals::new(from), total: 0., samples: 0 })
         });
         ctx.functions.insert("mean".into(), f);
@@ -1046,8 +1018,7 @@ impl<C: Ctx, E> Apply<C, E> for Mean {
         event: &Event<E>,
     ) -> Option<Value> {
         if self.from.update(ctx, from, event) {
-            // CR estokes: Is this correct? Think about it some more.
-            for v in &self.from.0 {
+            for v in self.from.flat_iter() {
                 if let Some(v) = v {
                     if let Ok(v) = v.clone().cast_to::<f64>() {
                         self.total += v;
@@ -1066,7 +1037,7 @@ pub(crate) struct Uniq(Option<Value>);
 
 impl<C: Ctx, E> Register<C, E> for Uniq {
     fn register(ctx: &mut ExecCtx<C, E>) {
-        let f: InitFn<C, E> = Arc::new(|_, from, _| {
+        let f: InitFn<C, E> = Arc::new(|_, from, _, _| {
             let mut t = Uniq(None);
             match from {
                 [e] => t.0 = e.current(),
@@ -1148,7 +1119,7 @@ pub struct Store {
 
 impl<C: Ctx, E> Register<C, E> for Store {
     fn register(ctx: &mut ExecCtx<C, E>) {
-        let f: InitFn<C, E> = Arc::new(|ctx, from, top_id| {
+        let f: InitFn<C, E> = Arc::new(|ctx, from, _, top_id| {
             let mut t = Store { queued: Vec::new(), dv: None, invalid: false, top_id };
             match from {
                 [to, val] => t.set(ctx, to.current(), val.current()),
@@ -1229,20 +1200,31 @@ impl Store {
                     dv.write(v);
                 }
             },
-            (Some(p), val) => {
-                let path = Path::from(p);
-                let dv = ctx.user.durable_subscribe(
-                    UpdatesFlags::empty(),
-                    path.clone(),
-                    self.top_id,
-                );
+            (Some(path), val) => {
                 if let Some(v) = val {
                     self.queue(v)
                 }
-                for v in self.queued.drain(..) {
-                    dv.write(v);
+                match &self.dv {
+                    Some((cur, dv)) if &path == cur => {
+                        for v in self.queued.drain(..) {
+                            dv.write(v);
+                        }
+                    }
+                    None | Some((_, _)) => {
+                        if let Some((cur, dv)) = self.dv.take() {
+                            ctx.user.unsubscribe(cur, dv, self.top_id);
+                        }
+                        let dv = ctx.user.durable_subscribe(
+                            UpdatesFlags::empty(),
+                            path.clone(),
+                            self.top_id,
+                        );
+                        for v in self.queued.drain(..) {
+                            dv.write(v);
+                        }
+                        self.dv = Some((path, dv));
+                    }
                 }
-                self.dv = Some((path, dv));
             }
         }
     }
@@ -1274,31 +1256,42 @@ fn varname(invalid: &mut bool, name: Option<Value>) -> Option<Chars> {
     }
 }
 
-pub struct StoreVar {
+pub struct Set {
     queued: Vec<Value>,
+    local: bool,
+    scope: Path,
     name: Option<Chars>,
     invalid: bool,
 }
 
-impl<C: Ctx, E> Register<C, E> for StoreVar {
+impl<C: Ctx, E> Register<C, E> for Set {
     fn register(ctx: &mut ExecCtx<C, E>) {
-        let f: InitFn<C, E> = Arc::new(|ctx, from, _| {
-            let mut t = StoreVar { queued: Vec::new(), name: None, invalid: false };
-            match from {
-                [name, value] => t.set(ctx, name.current(), value.current()),
-                _ => t.invalid = true,
-            }
-            Box::new(t)
-        });
-        ctx.functions.insert("store_var".into(), f);
+        let f = |local| -> InitFn<C, E> {
+            Arc::new(move |ctx, from, scope, _| {
+                let mut t = Set {
+                    queued: Vec::new(),
+                    local,
+                    scope,
+                    name: None,
+                    invalid: false,
+                };
+                match from {
+                    [name, value] => t.set(ctx, name.current(), value.current()),
+                    _ => t.invalid = true,
+                }
+                Box::new(t)
+            })
+        };
+        ctx.functions.insert("global_set".into(), f(false));
+        ctx.functions.insert("set".into(), f(true));
     }
 }
 
-impl<C: Ctx, E> Apply<C, E> for StoreVar {
+impl<C: Ctx, E> Apply<C, E> for Set {
     fn current(&self) -> Option<Value> {
         if self.invalid {
             Some(Value::Error(Chars::from(
-                "store_var(name: string [a-z][a-z0-9_]+, value): expected 2 arguments",
+                "set(name: string [a-z][a-z0-9_]+, value): expected 2 arguments",
             )))
         } else {
             None
@@ -1344,7 +1337,7 @@ impl<C: Ctx, E> Apply<C, E> for StoreVar {
     }
 }
 
-impl StoreVar {
+impl Set {
     fn queue_set(&mut self, v: Value) {
         self.queued.push(v)
     }
@@ -1357,14 +1350,26 @@ impl StoreVar {
     ) {
         if let Some(name) = varname(&mut self.invalid, name) {
             for v in self.queued.drain(..) {
-                ctx.user.set_var(&mut ctx.variables, name.clone(), v)
+                ctx.user.set_var(
+                    &mut ctx.variables,
+                    self.local,
+                    self.scope.clone(),
+                    name.clone(),
+                    v,
+                )
             }
             self.name = Some(name);
         }
         if let Some(value) = value {
             match self.name.as_ref() {
                 None => self.queue_set(value),
-                Some(name) => ctx.user.set_var(&mut ctx.variables, name.clone(), value),
+                Some(name) => ctx.user.set_var(
+                    &mut ctx.variables,
+                    self.local,
+                    self.scope.clone(),
+                    name.clone(),
+                    value,
+                ),
             }
         }
     }
@@ -1378,6 +1383,7 @@ impl StoreVar {
 }
 
 pub(crate) struct Load {
+    path: Option<Path>,
     cur: Option<Dval>,
     top_id: ExprId,
     invalid: bool,
@@ -1385,8 +1391,8 @@ pub(crate) struct Load {
 
 impl<C: Ctx, E> Register<C, E> for Load {
     fn register(ctx: &mut ExecCtx<C, E>) {
-        let f: InitFn<C, E> = Arc::new(|ctx, from, top_id| {
-            let mut t = Load { cur: None, invalid: false, top_id };
+        let f: InitFn<C, E> = Arc::new(|ctx, from, _, top_id| {
+            let mut t = Load { path: None, cur: None, invalid: false, top_id };
             match from {
                 [path] => t.subscribe(ctx, path.current()),
                 _ => t.invalid = true,
@@ -1403,7 +1409,9 @@ impl<C: Ctx, E> Apply<C, E> for Load {
             Load::err()
         } else {
             self.cur.as_ref().and_then(|dv| match dv.last() {
-                subscriber::Event::Unsubscribed => Some(Value::Error(Chars::from("#LOST"))),
+                subscriber::Event::Unsubscribed => {
+                    Some(Value::Error(Chars::from("#LOST")))
+                }
                 subscriber::Event::Update(v) => Some(v),
             })
         }
@@ -1417,9 +1425,13 @@ impl<C: Ctx, E> Apply<C, E> for Load {
     ) -> Option<Value> {
         match from {
             [name] => {
-                let target = name.update(ctx, event);
-                let up = target.is_some();
-                self.subscribe(ctx, target);
+                let up = match name.update(ctx, event) {
+                    None => false,
+                    Some(target) => {
+                        self.subscribe(ctx, Some(target));
+                        true
+                    }
+                };
                 if self.invalid {
                     if up {
                         Load::err()
@@ -1428,7 +1440,9 @@ impl<C: Ctx, E> Apply<C, E> for Load {
                     }
                 } else {
                     self.cur.as_ref().and_then(|dv| match event {
-                        Event::Variable(_, _) | Event::Rpc(_, _) | Event::User(_) => None,
+                        Event::Variable(_, _, _) | Event::Rpc(_, _) | Event::User(_) => {
+                            None
+                        }
                         Event::Netidx(id, value) if dv.id() == *id => Some(value.clone()),
                         Event::Netidx(_, _) => None,
                     })
@@ -1453,11 +1467,17 @@ impl<C: Ctx, E> Apply<C, E> for Load {
 impl Load {
     fn subscribe<C: Ctx, E>(&mut self, ctx: &mut ExecCtx<C, E>, name: Option<Value>) {
         if let Some(path) = pathname(&mut self.invalid, name) {
-            self.cur = Some(ctx.user.durable_subscribe(
-                UpdatesFlags::BEGIN_WITH_LAST,
-                path,
-                self.top_id,
-            ));
+            if Some(&path) != self.path.as_ref() {
+                if let (Some(path), Some(dv)) = (self.path.take(), self.cur.take()) {
+                    ctx.user.unsubscribe(path, dv, self.top_id);
+                }
+                self.path = Some(path.clone());
+                self.cur = Some(ctx.user.durable_subscribe(
+                    UpdatesFlags::BEGIN_WITH_LAST,
+                    path,
+                    self.top_id,
+                ));
+            }
         }
     }
 
@@ -1468,33 +1488,39 @@ impl Load {
     }
 }
 
-pub struct LoadVar {
+struct BoundVar {
+    scope: Path,
+    value: Value,
+}
+
+pub struct Get {
+    scope: Path,
     name: Option<Chars>,
-    cur: Option<Value>,
+    var: Option<BoundVar>,
     top_id: ExprId,
     invalid: bool,
 }
 
-impl<C: Ctx, E> Register<C, E> for LoadVar {
+impl<C: Ctx, E> Register<C, E> for Get {
     fn register(ctx: &mut ExecCtx<C, E>) {
-        let f: InitFn<C, E> = Arc::new(|ctx, from, top_id| {
-            let mut t = LoadVar { name: None, cur: None, invalid: false, top_id };
+        let f: InitFn<C, E> = Arc::new(|ctx, from, scope, top_id| {
+            let mut t = Get { scope, name: None, var: None, invalid: false, top_id };
             match from {
                 [name] => t.subscribe(ctx, name.current()),
                 _ => t.invalid = true,
             }
             Box::new(t)
         });
-        ctx.functions.insert("load_var".into(), f);
+        ctx.functions.insert("get".into(), f);
     }
 }
 
-impl<C: Ctx, E> Apply<C, E> for LoadVar {
+impl<C: Ctx, E> Apply<C, E> for Get {
     fn current(&self) -> Option<Value> {
         if self.invalid {
-            LoadVar::err()
+            Get::err()
         } else {
-            self.cur.to_owned()
+            self.var.as_ref().map(|bv| bv.value.clone())
         }
     }
 
@@ -1506,27 +1532,44 @@ impl<C: Ctx, E> Apply<C, E> for LoadVar {
     ) -> Option<Value> {
         match from {
             [name] => {
-                let target = name.update(ctx, event);
+                let up = match name.update(ctx, event) {
+                    None => false,
+                    Some(target) => {
+                        self.subscribe(ctx, Some(target));
+                        true
+                    }
+                };
                 if self.invalid {
-                    if target.is_some() {
-                        LoadVar::err()
+                    if up {
+                        Get::err()
                     } else {
                         None
                     }
                 } else {
-                    if let Some(target) = target {
-                        self.subscribe(ctx, Some(target));
-                    }
-                    match (self.name.as_ref(), event) {
-                        (None, _)
-                        | (Some(_), Event::Netidx(_, _))
-                        | (Some(_), Event::User(_))
-                        | (Some(_), Event::Rpc(_, _)) => None,
-                        (Some(vn), Event::Variable(tn, v)) if vn == tn => {
-                            self.cur = Some(v.clone());
+                    match (self.name.as_ref(), self.var.as_mut(), event) {
+                        (Some(vn), Some(bv), Event::Variable(es, en, v))
+                            if vn == en
+                                && Path::is_parent(es, &self.scope)
+                                && Path::is_parent(&bv.scope, es) =>
+                        {
+                            if &bv.scope != es {
+                                bv.scope = es.clone();
+                            }
+                            bv.value = v.clone();
                             Some(v.clone())
                         }
-                        (Some(_), Event::Variable(_, _)) => None,
+                        (Some(vn), None, Event::Variable(es, en, v))
+                            if vn == en && Path::is_parent(es, &self.scope) =>
+                        {
+                            self.var =
+                                Some(BoundVar { scope: es.clone(), value: v.clone() });
+                            Some(v.clone())
+                        }
+                        (None, _, _)
+                        | (Some(_), _, Event::Netidx(_, _))
+                        | (Some(_), _, Event::User(_))
+                        | (Some(_), _, Event::Rpc(_, _))
+                        | (Some(_), _, Event::Variable(_, _, _)) => None,
                     }
                 }
             }
@@ -1537,7 +1580,7 @@ impl<C: Ctx, E> Apply<C, E> for LoadVar {
                 }
                 self.invalid = true;
                 if up {
-                    LoadVar::err()
+                    Get::err()
                 } else {
                     None
                 }
@@ -1546,7 +1589,7 @@ impl<C: Ctx, E> Apply<C, E> for LoadVar {
     }
 }
 
-impl LoadVar {
+impl Get {
     fn err() -> Option<Value> {
         Some(Value::Error(Chars::from(
             "load_var(expr: variable name): expected 1 variable name as argument",
@@ -1554,10 +1597,25 @@ impl LoadVar {
     }
 
     fn subscribe<C: Ctx, E>(&mut self, ctx: &mut ExecCtx<C, E>, name: Option<Value>) {
-        if let Some(name) = varname(&mut self.invalid, name) {
-            self.cur = ctx.variables.get(&name).cloned();
-            ctx.user.ref_var(name.clone(), self.top_id);
-            self.name = Some(name);
+        match varname(&mut self.invalid, name) {
+            None => {
+                self.var = None;
+                self.name = None;
+                if let Some(old) = self.name.take() {
+                    ctx.user.unref_var(old.clone(), self.scope.clone(), self.top_id);
+                }
+            }
+            Some(name) if self.name.as_ref() == Some(&name) => (),
+            Some(name) => {
+                if let Some(old) = self.name.take() {
+                    ctx.user.unref_var(old.clone(), self.scope.clone(), self.top_id);
+                }
+                self.var = ctx.lookup_var(&self.scope, &name).map(|(scope, value)| {
+                    BoundVar { scope: scope.clone(), value: value.clone() }
+                });
+                ctx.user.ref_var(name.clone(), self.scope.clone(), self.top_id);
+                self.name = Some(name);
+            }
         }
     }
 }
@@ -1573,7 +1631,7 @@ pub(crate) struct RpcCall {
 
 impl<C: Ctx, E> Register<C, E> for RpcCall {
     fn register(ctx: &mut ExecCtx<C, E>) {
-        let f: InitFn<C, E> = Arc::new(|ctx, from, top_id| {
+        let f: InitFn<C, E> = Arc::new(|ctx, from, _, top_id| {
             let mut t = RpcCall {
                 args: CachedVals::new(match from {
                     [_, rest @ ..] => rest,
